#include <sys/keycodes.h>

#include <assert.h>
#include <display.h>
#include <err.h>
#include <errno.h>
#include <pixel.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <timespec.h>

#define STAGES 2
enum game_mode { TITLESCREEN, STAGE };
enum game_mode game_mode = TITLESCREEN;
bool stages_beat[STAGES];

#define PLAYFIELD_SIDE 100
#define TILE_SIDE 10
#define TILES (PLAYFIELD_SIDE / TILE_SIDE)
enum palette {
	BG = 0,
	TEXT,
	GRASS_TOP,
	GRASS_BROWN,
	GRASS_DARK,
	FLAG_STEM,
	FLAG_FLAG,
	STAR_STAR,
	ERROR,
};
static enum palette playfield[PLAYFIELD_SIDE * PLAYFIELD_SIDE];

#include "grass_tile.inc"
#include "dirt_tile.inc"
#include "flag_tile.inc"

#include "stage1_tile.inc"
#include "stage2_tile.inc"
#include "star_tile.inc"

static char tilemap[] = "          "
                        "          "
                        "          "
                        "!  g      "
                        "gg    g   "
                        "         g"
                        "        g "
                        "     g    "
                        "@   gdg   "
                        "gg  dddgg ";

static uint32_t main_window = 0;
static uint32_t window_width = 600;
static uint32_t window_height = 500;

static bool game_running = true;

static void on_disconnect(void *context) {
	(void) context;
	game_running = false;
}

static bool left_held, right_held, jump_held;

static double jump_maxheight = 21;
static double jump_duration = 0.4;

static void on_keyboard(void *context, uint32_t window, uint32_t codepoint) {
	(void) context;
	if (window != main_window)
		return;
	int key = KBKEY_DECODE(codepoint);
	switch (key) {
		// TODO: rebinding
		case KBKEY_LEFT: left_held = true; break;
		case -KBKEY_LEFT: left_held = false; break;
		case KBKEY_RIGHT: right_held = true; break;
		case -KBKEY_RIGHT: right_held = false; break;
		case KBKEY_SPACE: jump_held = true; break;
		case -KBKEY_SPACE: jump_held = false; break;
	}
}

static void on_quit(void *context, uint32_t window) {
	(void) context;
	if (window != main_window)
		return;
	game_running = false;
}

static void on_resize(void *context, uint32_t window, uint32_t width, uint32_t height) {
	(void) context;
	if (window != main_window)
		return;
	window_width = width;
	window_height = height;
}

static double timespec2double(struct timespec ts) {
	return ts.tv_sec + ts.tv_nsec / 1000.0 / 1000.0 / 1000.0;
}

static char sample_tilemap(size_t x, size_t y) {
	size_t tx = x / TILE_SIDE;
	size_t ty = y / TILE_SIDE;
	if (tx >= TILES || ty >= TILES) return ' ';
	return tilemap[ty * TILES + tx];
}

static bool solid_tile(size_t x, size_t y) {
	char tile = sample_tilemap(x, y);
	return tile != ' ' && tile != '!' && tile != '@';
}

static bool jumping = false;
static bool on_ground = false;
static double jump_start_y;
static double in_jump;
static double player_x, player_y;
static double player_dx, player_dy;

static void initialize_stage(void) {
	jumping = false;
	on_ground = false;
	player_dx = 0;
	player_x = 0;
	player_y = 0;
	for (size_t ty = 0; ty < TILES; ty++) {
		for (size_t tx = 0; tx < TILES; tx++) {
			if (tilemap[ty * TILES + tx] == '@') {
				player_x = tx * TILE_SIDE;
				player_y = ty * TILE_SIDE;
			}
		}
	}
}

static size_t selection_index = 0;
static bool keys_released;
static bool stage_selected; 

static void initialize(void) {
	switch (game_mode) {
		case TITLESCREEN: keys_released = false; stage_selected = false; break;
		case STAGE: initialize_stage(); break;
		default: printf("%i\n", game_mode); break;
	}
}

static void update_stage(struct timespec now, struct timespec dt_timespec) {
	(void) now;
	double dt = timespec2double(dt_timespec);
	player_x += player_dx * dt;

	if (left_held) player_dx = -30;
	else if (right_held) player_dx = 30;
	else player_dx = 0;

	if (on_ground && jump_held && !jumping) {
		jumping = true;
		in_jump = 0;
		jump_start_y = player_y;
	}
	if (!on_ground && !jumping) {
		jumping = true;
		in_jump = jump_duration;
		jump_start_y = player_y + jump_maxheight;
	}
	if (jumping) {
		in_jump += dt;
		// Jump is a quadratic arc
		// h = at^2 + bt + c
		//
		// Beginning of jump, zero height
		// t = 0, h = 0
		// 0 = a0^2 + b0 + c
		// c = 0
		//
		// At 2*duration, zero height
		// t = 2T, h = 0
		// 0 = a(2T)^2 + b(2T) + 0
		// 0 = 4aT^2 + 2bT
		// 0 = 2aT^2 + bT
		// 0 = 2aT + b
		// b = -2aT
		//
		// At duration, maximum height
		// t = T, h = H
		// H = aT^2 + bT + 0
		// H = aT^2 - 2aT^2
		// H = -aT^2
		// a = -H / (T^2)
		// b
		double a = -jump_maxheight / (jump_duration * jump_duration);
		double b = -2 * a * jump_duration;
		double jump_height = a * in_jump * in_jump + b * in_jump;
		player_y = jump_start_y - jump_height;
		// player_dy is negative of derivative of jump_height
		// H = a*t^2 + b*t
		// H' = 2*a*t + b
		player_dy = -2 * a * in_jump - b;
	}

	// Collision against walls of the map
	if (player_x < 0) {
		player_x = 0;
		player_dx = 0;
	} else if (player_x > PLAYFIELD_SIDE - TILE_SIDE) {
		player_x = PLAYFIELD_SIDE - TILE_SIDE;
		player_dx = 0;
	}
	if (player_y < 0) {
		player_y = 0;
		player_dy = 0;
	} else if (player_y > PLAYFIELD_SIDE - TILE_SIDE) {
		player_y = PLAYFIELD_SIDE - TILE_SIDE;
		player_dy = 0;
		game_mode = TITLESCREEN;
		initialize();
		return;
	}

	// Collision against tiles
	size_t px = player_x;
	size_t py = player_y;
	size_t left_cx = px - 1;
	size_t right_cx = px + TILE_SIDE;
	size_t leftright_top_cy = py;
	size_t leftright_bottom_cy = py + TILE_SIDE - 3; //XXX: proper handling of snapping
	if ((solid_tile(left_cx, leftright_top_cy) ||
	     solid_tile(left_cx, leftright_bottom_cy)) &&
		player_dx < 0) {
		size_t snapped_px = (px + TILE_SIDE - 1) / TILE_SIDE * TILE_SIDE;
		player_x = snapped_px;
		player_dx = 0;
	}
	if ((solid_tile(right_cx, leftright_top_cy) ||
	     solid_tile(right_cx, leftright_bottom_cy)) &&
	    player_dx > 0) {
		size_t snapped_px = px / TILE_SIDE * TILE_SIDE;
		player_x = snapped_px;
		player_dx = 0;
	}

	px = player_x;
	py = player_y;
	size_t topbottom_left_cx = px;
	size_t topbottom_right_cx = px + TILE_SIDE - 1;
	size_t top_cy = py - 1;
	size_t bottom_cy = py + TILE_SIDE;

	on_ground = false;
	if (solid_tile(topbottom_left_cx, bottom_cy) ||
	    solid_tile(topbottom_right_cx, bottom_cy)) {
		on_ground = true;
		if (player_dy > 0) {
			jumping = false;
			size_t snapped_py = py / TILE_SIDE * TILE_SIDE;
			player_y = snapped_py;
		}
	}
	if ((solid_tile(topbottom_left_cx, top_cy) ||
	     solid_tile(topbottom_right_cx, top_cy)) &&
		player_dy < 0) {
		size_t snapped_py = (py + TILE_SIDE - 1) / TILE_SIDE * TILE_SIDE;
		player_y = snapped_py;
		in_jump = jump_duration;
		jump_start_y = player_y + jump_maxheight;
	}

	// Touching tiles
	size_t midpoint_x = player_x + TILE_SIDE / 2;
	size_t midpoint_y = player_y + TILE_SIDE / 2;
	if (sample_tilemap(midpoint_x, midpoint_y) == '!') {
		stages_beat[selection_index] = true;
		if (selection_index < STAGES - 1)
			selection_index++;
		game_mode = TITLESCREEN;
		initialize();
		return;		
	}
}

static void update_titlescreen(struct timespec now, struct timespec dt_timespec) {
	(void) now;
	(void) dt_timespec;
	if (keys_released && left_held && selection_index > 0)
		selection_index--;
	if (keys_released && right_held && selection_index < STAGES - 1)
		selection_index++;
	if (keys_released && jump_held)
		stage_selected = true;
	if (!left_held && !right_held && !jump_held)
		keys_released = true;
	else
		keys_released = false;
	if (stage_selected && keys_released) {
		game_mode = STAGE;
		initialize();
	}
}

static void update(struct timespec now, struct timespec dt_timespec) {
	switch (game_mode) {
		case TITLESCREEN: update_titlescreen(now, dt_timespec); break;
		case STAGE: update_stage(now, dt_timespec); break;
		default: printf("%i\n", game_mode); break;
	}
}

static void draw_tile(enum palette tile[], size_t x, size_t y) {
	assert(x + TILE_SIDE <= PLAYFIELD_SIDE);
	assert(y + TILE_SIDE <= PLAYFIELD_SIDE);
	for (size_t ty = 0; ty < TILE_SIDE; ty++) {
		for (size_t tx = 0; tx < TILE_SIDE; tx++) {
			enum palette pixel = tile[ty * TILE_SIDE + tx];
			playfield[(y + ty) * PLAYFIELD_SIDE + x + tx] = pixel;
		}
	}
}

static void draw_tiles(void) {
	assert(strlen(tilemap) == TILES * TILES);
	for (size_t ty = 0; ty < TILES; ty++) {
		for (size_t tx = 0; tx < TILES; tx++) {
			enum palette *tile = NULL;
			switch(tilemap[ty * TILES + tx]) {
				case 'g': tile = grass_tile; break;
				case 'd': tile = dirt_tile; break;
				case '!': tile = flag_tile; break;
			}
			if (tile)
				draw_tile(tile, tx * TILE_SIDE, ty * TILE_SIDE);
		}
	}
}

static void draw_stage(void) {
	memset(playfield, 0, sizeof(playfield));
	draw_tiles();
	draw_tile(grass_tile, player_x, player_y);
}

static void draw_titlescreen(void) {
	memset(playfield, 0, sizeof(playfield));
	for (size_t i = 0; i < STAGES; i++) {
		size_t x = i * TILE_SIDE + TILE_SIDE;
		size_t y = 0;
		if (stages_beat[i])
			draw_tile(star_tile, x, y);
	}
	draw_tile(stage1_tile, 1 * TILE_SIDE, TILE_SIDE);
	draw_tile(stage2_tile, 2 * TILE_SIDE, TILE_SIDE);
	size_t underline_y = 2 * TILE_SIDE + 1;
	size_t underline_x = selection_index * TILE_SIDE + TILE_SIDE + 2;
	for (size_t x = 0; x < TILE_SIDE - 4; x++)
		playfield[underline_y * PLAYFIELD_SIDE + underline_x + x] = TEXT;
}

static void draw(void) {
	switch (game_mode) {
		case TITLESCREEN: draw_titlescreen(); break;
		case STAGE: draw_stage(); break;
		default: printf("%i\n", game_mode); break;
	}
}

int main(int argc, char *argv[]) {
	struct display_connection *connection = display_connect_default();
	if (!connection && errno == ECONNREFUSED)
		display_spawn(argc, argv);
	if (!connection)
		err(1, "Could not connect to display server");

	display_create_window(connection, main_window);
	display_resize_window(connection, main_window, window_width, window_height);
	display_show_window(connection, main_window);

	struct display_event_handlers handlers = {0};
	handlers.disconnect_handler = on_disconnect;
	handlers.keyboard_handler = on_keyboard;
	handlers.quit_handler = on_quit;
	handlers.resize_handler = on_resize;

	uint32_t palette_colours[] = {
		[BG] = make_color(128, 128, 255),
		[TEXT] = make_color(255, 255, 255),
		[GRASS_TOP] = make_color(0, 220, 0),
		[GRASS_BROWN] = make_color(200, 128, 0),
		[GRASS_DARK] = make_color(100, 64, 0),
		[FLAG_STEM] = make_color(0, 0, 0),
		[FLAG_FLAG] = make_color(255, 0, 0),
		[STAR_STAR] = make_color(255, 255, 0),
		[ERROR] = make_color(255, 0, 255),
	};

	long fps_cap = 100;
	struct timespec min_dt = timespec_make(0, 1000L * 1000L * 1000L / fps_cap);
	struct timespec last_frame;
	clock_gettime(CLOCK_MONOTONIC, &last_frame);

	initialize();

	int fps = 0;
	while (game_running) {
		struct timespec now;
		clock_gettime(CLOCK_MONOTONIC, &now);
		struct timespec dt = timespec_sub(now, last_frame);
		if (timespec_lt(dt, min_dt)) {
			struct timespec sleep = timespec_sub(min_dt, dt);
			nanosleep(&sleep, NULL);
			continue;
		}

		if (last_frame.tv_sec != now.tv_sec) {
			char *title = NULL;
			asprintf(&title, "%i fps", fps);
			display_title_window(connection, main_window, title);
			free(title);
			fps = 0;
		} else
			fps++;

		update(now, dt);
		draw();

		size_t pixels;
		if (__builtin_mul_overflow(window_width, window_height, &pixels))
			errx(1, "Window size too big");
		uint32_t *framebuffer = calloc(pixels, sizeof(uint32_t));

		uint32_t shorter_side = window_height < window_width ? window_height : window_width;
		size_t y_offset = (window_height - shorter_side) / 2;
		size_t x_offset = (window_width - shorter_side) / 2;
		for (size_t y = 0; y < window_height; y++) {
			for (size_t x = 0; x < window_width; x++) {
				if (y < y_offset || x < x_offset ||
				    y >= y_offset + shorter_side || x >= x_offset + shorter_side)
					framebuffer[y * window_width + x] = make_color_a(0, 0, 0, 200);
				else {
					size_t pf_x = (x - x_offset) * PLAYFIELD_SIDE / shorter_side;
					size_t pf_y = (y - y_offset) * PLAYFIELD_SIDE / shorter_side;
					uint32_t colour = palette_colours[playfield[pf_y * PLAYFIELD_SIDE + pf_x]];
					framebuffer[y * window_width + x] = colour;
				}
			}
		}
		display_render_window(connection, main_window, 0, 0, window_width, window_height, framebuffer);
		free(framebuffer);

		while (!display_poll_event(connection, &handlers));
		
		last_frame = now;
	}

	display_disconnect(connection);

	return 0;
}
